#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import shutil
import pickle
#from multiprocessing import Process
import multiprocessing as multiproc

from cortix.src.module import Module
from cortix.src.port import Port

class Network:
    """Cortix network.

    Attributes
    ----------

    num_networks: int
        Number of instances of this class.
    """

    num_networks = 0

    def __init__(self):
        """Module super class constructor.

        Attributes
        ----------
          max_n_modules_for_data_copy_on_root: int
              When using MPI the `network` will copy the data from all modules on the
              root process. This can generate an `out of memory` condition. This variable
              sets the maximum number of processes for which the data will be copied.
              Default is 1000.
       """

        self.id = Network.num_networks

        self.name = 'network-'+str(self.id)
        self.log = None

        self.max_n_modules_for_data_copy_on_root = 1000

        self.modules = list()

        self.gv_edges = list()
        self.gv_info = 'undirectional'

        self.use_mpi = None
        self.use_multiprocessing = None
        self.is_multiproc_start_method_set = False

        self.rank = None
        self.size = None
        self.comm = None

        self.save = False # save all network modules

        Network.num_networks += 1

    def module(self, m):
        """Add a module.
        """

        assert isinstance(m, Module), 'm must be a module'

        if m not in self.modules:
            m.use_mpi = self.use_mpi
            m.use_multiprocessing = self.use_multiprocessing
            self.modules.append(m)
            m.id = len(self.modules)-1  # see module doc for module id
            if not m.name:
             m.name = m.__class__.__name__
            m.log = self.log

    def add_module(self, m):
        """Alternative name to `module()`.
        """

        self.module(m)

    def connect(self, module_port_a, module_port_b, info=None):
        """Connect two modules using either their ports directly or inferred ports.

        A connection always opens a channel for data communication in both ways.
        That is, both sends and receives are allowed.

        Note
        ----
        The simplest form of usage is with arguments: (`module_a`, `module_b`).
        In this case, a `port` with the name of `module_a` **must** exist in `module_b`,
        and vice-versa (port names as `str` in lower case). In addition, the connect
        must not be called again with these same two modules, else the underlying
        connection will be overriden.

        For more rigorous connection, the user is advised to fully specify the
        module and the port in each list argument.

        Parameters
        ----------
        module_port_a: list([Module,Port]) or list([Module,str]) or Module
            First `module`-`port` to connect.

        module_port_b: list([Module,Port]) or list([Module,str]) or Module
            Second `module`-`port` to connect.

        info: str
            Information on the directionality of the information flow. This is for
            graph visualization purposes only. The default value will use the order
            in the argument list to define the direction. Default: None. If set
            to `bidiretional`, will create an edge with arrows on both ends in the graph figure.
            if set to 'undirectional' will create a plain edge lines. If set to 'directional' will
            create edges with the arrow pointing in one direction dictated by the edge ordering.
            If left as the default, None, a undirected edge will be drawn which means bidirectionality.
        """

        if info:
            assert isinstance(info, str)
            assert info in ['undirectional', 'directional', 'bidirectional']
            self.gv_info = info

        if isinstance(module_port_a, Module) and isinstance(module_port_b, Module):

            module_a = module_port_a
            module_b = module_port_b

            assert module_a.name and module_b.name  # sanity check

            assert module_a in self.modules, 'module %r not in network.'%module_a.name
            assert module_b in self.modules, 'module %r not in network.'%module_b.name

            # Connect ports
            port_a = module_a.get_port(module_b.name.lower())
            port_b = module_b.get_port(module_a.name.lower())

            port_a.connect(port_b)

            # Record connectivity for graph viz.
            idx_a = self.modules.index(module_a)
            idx_b = self.modules.index(module_b)

            if (str(idx_a), str(idx_b), info) not in self.gv_edges:
                self.gv_edges.append((str(idx_a), str(idx_b), info))

                # Double edges for bidiectional: deprecated
                #if self.gv_info == 'bidirectional' and (str(idx_b), str(idx_a)) not in self.gv_edges:
                #    self.gv_edges.append((str(idx_b), str(idx_a)))

        elif isinstance(module_port_a, list) and isinstance(module_port_b, list):

            assert len(module_port_a) == 2 and isinstance(module_port_a[0], Module) and\
                  (isinstance(module_port_a[1], str) or isinstance(module_port_a[1], Port))

            assert len(module_port_b) == 2 and isinstance(module_port_b[0], Module) and\
                  (isinstance(module_port_b[1], str) or isinstance(module_port_b[1], Port))

            module_a = module_port_a[0]
            module_b = module_port_b[0]

            assert module_a.name and module_b.name  # sanity check

            assert module_a in self.modules, 'module %r not in network.'%module_a.name
            assert module_b in self.modules, 'module %r not in network.'%module_b.name

            idx_a = self.modules.index(module_a)
            idx_b = self.modules.index(module_b)

            self.gv_edges.append((str(idx_a), str(idx_b), info))

            # Double edges for bidirectional: deprecated
            #if self.gv_info == 'bidirectional':
            #    self.gv_edges.append((str(idx_b), str(idx_a)))

            if isinstance(module_port_a[1], str):
                port_a = module_a.get_port(module_port_a[1])
            elif isinstance(module_port_a[1], Port):
                port_a = module_port_a[1]
            else:
                assert False, 'help!'

            if isinstance(module_port_b[1], str):
                port_b = module_b.get_port(module_port_b[1])
            elif isinstance(module_port_b[1], Port):
                port_b = module_port_b[1]
            else:
                assert False, 'help!'

            port_a.connect(port_b)

        else:
            assert False, ' not implemented.'

        return

    def __run(self, save=False):
        """
        Internal method to run the network simulation. Do not use this method, it is
        intended for Cortix to run it.

        This function concurrently executes the `cortix.src.module.run` function
        for each module in the network. Modules are run using either MPI or
        Python Multiprocessing, depending on the user configuration. This is `not` a multi-threaded 
        application, it will always start multiple processes, either in MPI or Python Multiprocessing.

        Note
        ----
        When using multiprocessing, data from the modules state are copied to the master
        process after the `__run()` method of the modules is finished.
        """
        assert len(self.modules) >= 1, 'the network must have a list of modules.'

        # Create directory for saving modules states
        if self.rank == 0 or self.use_multiprocessing:
            shutil.rmtree('.ctx-saved', ignore_errors=True)
            os.makedirs('.ctx-saved')

        # Running under MPI
        #------------------
        if self.use_mpi:

            # Synchronize in the beginning
            assert self.size == len(self.modules) + 1,\
                'Incorrect number of processes (Required %r, got %r)'%\
                (len(self.modules) + 1, self.size)
            self.comm.Barrier()

            # Assign an mpi rank to all ports of a module using the module list index
            # If a port has rank assignment from a previous run; leave it alone
            for mod in self.modules:
                rank = self.modules.index(mod)+1
                for port in mod.ports:
                    if port.rank is None:
                        port.rank = rank

            # Assign a unique port id to all ports
            # If a port has id assignment from a previous run; leave it alone
            i = 0
            for mod in self.modules:
                for port in mod.ports:
                    port.use_mpi = self.use_mpi
                    if port.id is None:
                        port.id = i
                        i += 1

            # Parallel run module in MPI
            if self.rank != 0:
                mod = self.modules[self.rank-1]
                self.log.info('Launching Module {}'.format(mod))
                mod.run_and_save()

            # Sync here at the end
            self.comm.Barrier()

        # Running under Python multiprocessing
        #-------------------------------------
        else:

            # Parallel run all modules in Python multiprocessing
            if not self.is_multiproc_start_method_set:
                multiproc.set_start_method('spawn') # fresh interpreter for child processes
                self.is_multiproc_start_method_set = True
            processes = list()

            for mod in self.modules:
                self.log.info('Launching Module {}'.format(mod))
                # Note: on the other end, args will arrive as a doubly tuple: ((self.log,),)
                proc = multiproc.Process(target=mod.run_and_save, args=(self.log,))
                #proc = multiproc.Process(target=mod.run_and_save, args=(self.log,), kwargs={'logger':self.log})
                processes.append(proc)
                proc.start()

            # Synchronize at the end
            for proc in processes:
                proc.join()

        # Reload saved modules
        #---------------------
        if self.use_mpi:
            # Make double sure all processes are in sync here before reading files from disk
            self.comm.Barrier()

        num_files = 0
        for file_name in os.listdir('.ctx-saved'):

            self.gv_edges = list() # re-initialize since ports were not pickled

            if file_name.endswith('.pkl'):
                num_files += 1
                file_name = os.path.join('.ctx-saved', file_name)
                with open(file_name, 'rb') as fin:
                    module = pickle.load(fin)
                    # Reintroduce logging
                    module.log = self.log
                    self.modules[module.id] = module

        if num_files and num_files != len(self.modules):
            self.log.warning('Network::run(): not all modules reloaded from disk.\
                              # modules = %i; # files = %i'%(len(self.modules), num_files))

        if self.use_mpi:
            # Make double sure all are in sync here before going forward
            # this solves the problem of processes running behind reading files
            # that do not exist anymore
            self.comm.Barrier()

    def draw(self, graph_attr=None, node_attr=None, engine='twopi', lr=False,
             size=None, ports=False, node_shape='hexagon'):
        """Build a `graphviz` graph and draw the network saving it to a file.

        Parameters
        ----------
        graph_attr: dict(str:str)
            Graph attributes per GraphViz library. Use a dictionary of key:value
            strings.

        node_attr: dict(str:str)
            Node attributes per GraphViz library. Use a dictionary of key:value
            strings.

        engine: str
            Name of drawing engine (from GraphViz):
            'dot': draws directed graphs. It works well on directed acyclic graphs
                   and other graphs that can be drawn as hierarchies or have a natural
                   flow.

            'neato': draws  undirected  graphs  using  a  spring model  and reducing
                     the  related  energy.

            'twopi': draws graphs using a radial layout.  Basically,one node is
                     chosen as the center and put at the origin. The remaining nodes
                     are placed on a sequence of concentric  circles  centered  about
                     the  origin,  each  a  fixed  radial  distance  from  the
                     previous circle.  Allnodes distance 1 from the center are placed
                     on the first circle; all nodes distance 1 from a nodeon the
                     first circle are placed on the second circle; and so forth.

            'circo': draws graphs using a circular layout. The tool identifies
                     biconnected components and draws the nodes of the component
                     on a circle. The block-cutpoint tree is then laid out using a
                     recursive radial algorithm. Edge crossings within acircle  are
                     minimized by placing as manyedges on the circle’s
                     perimeter as possible. In particular,ifthecomponent is
                     outer planar, the component will have a planar layout. If a node
                     belongs to multiple non-triv-ial biconnected components, the
                     layout puts the node in one of them. By default, this is the
                     first non-trivialcomponent found in the search from the root
                     component.

            'fdp': draws undirected graphs using a spring model. It relies on a
                   force-directed approach in the spirit of Fruchterman and Reingold.

            `sfdp` also  draws  undirected  graphs  using  the spring model described
                   above,but  it  uses  a  multi-scaleapproach to produce layouts of
                   large graphs in a reasonably short time.patchworkdraws the graph
                   as a squarified treemap. The clusters of the graph are used to
                   specify the tree.

            'osage' draws the graph using its cluster structure. For a givencluster,
                    each of its subclusters is laid out inter-nally.Then  the
                    subclusters, plus anyremaining  nodes, are repositioned based on
                    the  cluster’s

        lr: boolean
            True draws graph left to right. False draws top down.

        size: str
            Pair of integer numbers in a string: 'a,b'.

        ports: boolean
            Draw name of ports

        node_shape: str
            Select node shape to draw. Options:
            'box', 'polygon', 'ellipse', 'oval',
            'circle', 'point', 'egg', 'triangle',
            'plaintext', 'plain', 'diamond', 'trapezium',
            'parallelogram', 'house', 'pentagon', 'hexagon',
            'septagon', 'octagon', 'doublecircle', 'doubleoctagon',
            'tripleoctagon', 'invtriangle', 'invtrapezium', 'invhouse',
            'Mdiamod', 'Msquare', 'Mcircle', 'rect',
            'rectangle', 'square', 'star', 'none',
            'underline', 'cylinder', 'note', 'tab',
            'folder', 'box3d', 'component', 'promoter',
            'cds', 'terminator', 'utr', 'primersite',
            'restrictionsite', 'fivepoverhang', threepoverhang', 'noverhang',
            'rnastab', 'proteasesite', 'proteinstab', 'rpromoter',
            'rarrow', 'larrow', 'lpromoter'.
        """

        # Delete existing graph files if any.
        if os.path.isfile(self.name+'.gv'):
            os.remove(self.name+'.gv')
        if os.path.isfile(self.name+'.gv.png'):
            os.remove(self.name+'.gv.png')

        # Import here to avoid broken dependency. Only this method needs graphviz.
        if self.gv_info == 'undirectional':
            from graphviz import Graph
        else:
            from graphviz import Digraph

        if graph_attr is None:
            graph_attr = {'splines':'true', 'overlap':'scale', 'ranksep':'1.5'}

        if node_attr is None:
            node_attr = {'shape':node_shape, 'style':'filled'}

        if self.gv_info == 'undirectional':
            graph = Graph(name=self.name, comment='Network::draw:graphviz-graph', format='png',
                        graph_attr=graph_attr, node_attr=node_attr, engine=engine)
        else:
            graph = Digraph(name=self.name, comment='Network::draw:graphviz-digraph', format='png',
                          graph_attr=graph_attr, node_attr=node_attr, engine=engine)

        if lr:
            graph.attr(rankdir='LR')

        if size:
            graph.attr(size=size)

        #graph.attr('node', shape=node_shape)
        #graph.attr('edge', dir='none')

        # Create node ids and labels
        for idx, mod in enumerate(self.modules):
            graph.node(str(idx), mod.name)

        # Create edges dir= none, both, right, left
        for edg in self.gv_edges:
            if edg[2] == 'bidirectional':
                graph.edge(edg[0], edg[1], dir='both')
            else:
                graph.edge(edg[0], edg[1])

        graph.render()

        return graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import sys
import shutil
import logging
import time
import datetime

from cortix.src.network import Network

class Cortix:
    """Cortix main class definition.

    The typical Cortix run file workflow:

    1. Create the `Cortix` object
    1. Create a network object and attach to the Cortix object
    2. Populate the network singleton with modules
    3. Run and close `Cortix`

    """

    def __init__(self, use_mpi=False, splash=False, log_filename_stem='cortix'):
        """Construct a Cortix simulation object.

        Parameters
        ----------
        use_mpi: bool
            True for MPI, False for multiprocessing.
        splash: bool
            Show the Cortix splash image.

        Attributes
        ----------
        network: Network
             A network of modules and their connectivity.
        use_mpi: bool
            `True` for MPI, `False` for Multiprocessing.
        use_multiprocessing: bool
            `False` for MPI, `True` for Multiprocessing.
        splash: bool
            Show the Cortix splash image.
        comm: mpi4py.MPI.Intracomm
            MPI.COMM_WORLD (if using MPI else None).
        rank: int
            The current MPI rank (if using MPI else None).
        size: int
            size of the group associated with MPI.COMM_WORLD.

        """
        self.use_mpi = use_mpi
        self.use_multiprocessing = not use_mpi
        self.comm = None
        self.rank = None
        self.size = None

        self.splash = splash

        self.__network = None
        self.log_filename_stem = log_filename_stem
        self.logger_name = self.log_filename_stem
        # Fall back to multiprocessing if mpi4py is not available
        if self.use_mpi:
            try:
                from mpi4py import MPI
                self.comm = MPI.COMM_WORLD
                self.rank = self.comm.Get_rank()
                self.size = self.comm.size
            except ImportError:
                self.use_mpi = False

        # Setup the global logger
        self.__create_logger()

        # Wrap-up init
        if self.rank == 0 or self.use_multiprocessing:

            if self.splash:
                self.log.info('Created Cortix object %s', self.__get_splash(begin=True))
            else:
                self.log.info('Created Cortix object')

            # Initialize all date and timings
            self.wall_clock_time_start = time.time()
            self.wall_clock_time_end = self.wall_clock_time_start
            self.end_run_date = datetime.datetime.today().strftime('%d%b%y %H:%M:%S')

            # Delete any existing .ctx-saved/*
            shutil.rmtree('.ctx-saved', ignore_errors=True)

    def __set_network(self, n):
        assert isinstance(n, Network)
        n.use_mpi = self.use_mpi
        n.use_multiprocessing = self.use_multiprocessing
        n.rank = self.rank
        n.size = self.size
        n.comm = self.comm
        n.log  = self.log
        self.__network = n
    def __get_network(self):
        return self.__network
    network = property(__get_network, __set_network, None, None)

    def run(self, save=False):
        """Run the Cortix network simulation.
        """

        self.__network._Network__run(save)

        if self.rank == 0 or self.use_multiprocessing:
            self.wall_clock_time_end = time.time()
            self.log.info('run()::Elapsed wall clock time [s]: '+
                          str(round(self.wall_clock_time_end-self.wall_clock_time_start, 2)))

    def close(self):
        """Closes the cortix object properly before destruction.

        User is strongly advised to call this method at the end of the run file otherwise
        timings will not be recorded.

        """
        # Sync here before close
        if self.use_mpi:
            self.comm.Barrier()

        if self.rank == 0 or self.use_multiprocessing:

            if self.splash:
                self.log.info('Closed Cortix object.'+self.__get_splash(end=True))
            else:
                self.log.info('Closed Cortix object.')

            self.wall_clock_time_end = time.time()

            self.log.info('close()::Elapsed wall clock time [s]: '+
                          str(round(self.wall_clock_time_end-self.wall_clock_time_start, 2)))
            logging.shutdown()

        return

    def __create_logger(self):
        """A helper function to setup the logging facility.

        The Python logging module is used to create two handlers, namely one for stdout output and
        another for file output. File output in multiprocessing has its challenges. The implementation
        here is experimental. The main process creates the Python logging logger. The issue is how
        to share the logger with the Cortix modules that run in a separate process. If using MPI, the
        logger is not implemented and modules will not share the logger. If using Python Multiprocessing
        the logger is passed to the child process.

        Note: help(logging). Levels: 0 = NOTSET, 10 = DEBUG, 20 = INFO, 30 = WARNING,
              40 = ERROR, 50 = FATAL
        """

        # File removal
        if self.rank == 0 or self.use_multiprocessing:
            if os.path.isfile(self.log_filename_stem+'.log'):
                os.remove(self.log_filename_stem+'.log')

        # Sync here to allow for file removal
        if self.use_mpi:
            self.comm.Barrier()

        self.log = logging.getLogger(self.logger_name)
        self.log.setLevel(logging.DEBUG)

        # Create handlers
        if not self.log.hasHandlers():
            file_handler = logging.FileHandler(self.log_filename_stem+'.log')
            file_handler.setLevel(logging.DEBUG)

            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.DEBUG)
            console_handler.setStream(sys.stdout)

            # Formatter added to handlers
            if self.use_mpi:
                fs = '[rank:{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s'.format(self.rank)
            else:
                fs = "[{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s".format(os.getpid())

            formatter = logging.Formatter(fs)
            file_handler.setFormatter(formatter)
            console_handler.setFormatter(formatter)

            # Add handlers to logger; this creates a handlers list
            self.log.addHandler(file_handler)
            self.log.addHandler(console_handler)
        else:
            assert False, 'Fatal: Logger exists.'

        return

    def __get_splash(self, begin=None, end=None):
        '''Returns the Cortix splash logo.

        Note
        ----
        Call this internal method with one argument only. Either `begin=True` or
        `end=True`.

        Parameters
        ----------
        begin: bool
            True for the beginning message, false for the ending.

        Returns
        -------
        splash: str
            The Cortix splash logo.

        '''

        assert begin is None or end is None
        if begin:
            end = False
        elif end:
            begin = False


        splash = \
        '_____________________________________________________________________________\n'+\
        '      ...                                        s       .     (TAAG Fraktur)\n'+\
        '   xH88"`~ .x8X                                 :8      @88>\n'+\
        ' :8888   .f"8888Hf        u.      .u    .      .88      %8P      uL   ..\n'+\
        ':8888>  X8L  ^""`   ...ue888b   .d88B :@8c    :888ooo    .     .@88b  @88R\n'+\
        'X8888  X888h        888R Y888r ="8888f8888r -*8888888  .@88u  ""Y888k/"*P\n'+\
        '88888  !88888.      888R I888>   4888>"88"    8888    ''888E`    Y888L\n'+\
        '88888   %88888      888R I888>   4888> "      8888      888E      8888\n'+\
        '88888 `> `8888>     888R I888>   4888>        8888      888E      `888N\n'+\
        '`8888L %  ?888   ! u8888cJ888   .d888L .+    .8888Lu=   888E   .u./"888&\n'+\
        ' `8888  `-*""   /   "*888*P"    ^"8888*"     ^%888*     888&  d888" Y888*"\n'+\
        '   "888.      :"      "Y"          "Y"         "Y"      R888" ` "Y   Y"\n'+\
        '     `""***~"`                                           ""\n'+\
        '                             https://cortix.org                              \n'+\
        '_____________________________________________________________________________'

        if begin:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                             L A U N C H I N G                               \n'

        else:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                           T E R M I N A T I N G                             \n'

        return message + splash

    def __del__(self):
        """Destructs a Cortix simulation object.

        Warning
        -------
        By the time the body of this function is executed, the machinery of
        variables may have been deleted already. For example, `logging` is no longer
        there; do the least amount of work here.

        """

        pass

if __name__ == '__main__':
    c = Cortix()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from .src.cortix_main import Cortix

from .src.network import Network
from .src.module import Module
from .src.port import Port

from .support.units import Units
from .support.phase import Phase
from .support.quantity import Quantity
from .support.species import Species
from .support.chemeng.reaction_mechanism import ReactionMechanism
from .support.chemeng.reaction_mechanism import print_reaction_sub_mechanisms



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org
"""Cortix module.
   Convenient non-SI (and SI) units to be transformed to SI.

   Usage
   -----

   On guest code.py:
   from cortix import Units as unit

   examples:
   size = 10*unit.meter
   temp = 10*unit.kelvin

   temp = convert_temp(10, 'C', 'K')

   incorrect: temp = 20*unit.F
   correct:   temp = unit.convert_temperature(20, 'F', 'K')
"""

import scipy.constants as scipy_cte

class Units:

    # make scipy constants available
    scipy_cte = scipy_cte

    # all class data

    # unit prefix
    mega = scipy_cte.mega
    kilo = scipy_cte.kilo
    centi = scipy_cte.centi
    milli = scipy_cte.milli

    # time
    second = 1.0
    minute = scipy_cte.minute
    sec = second
    min = minute
    hour = scipy_cte.hour
    day = scipy_cte.day
    year = scipy_cte.year

    # event driven or iterative processes
    event = 1.0
    step = 1.0

    # number
    mole = 1.00

    # mass
    gram = scipy_cte.gram
    kg = kilo*gram
    lb = scipy_cte.lb

    # length
    meter = 1.0
    cm = centi*meter
    ft = scipy_cte.foot
    foot = ft
    inch = scipy_cte.inch

    # area
    barn = 1.0e-28 * meter**2 # nuclear cross section

    # volume
    cc = (centi*meter)**3
    liter = scipy_cte.liter
    L = liter
    mL = milli*L
    gallon = scipy_cte.gallon

    # energy/power/pressure
    joule = 1.0
    kj = kilo*joule
    watt = 1.0
    btu = scipy_cte.Btu
    Btu = btu
    pascal = 1.0
    bar = scipy_cte.bar
    psi = scipy_cte.psi
    atm = scipy_cte.atm

    # charge/electric potential/current
    coulomb = 1.0
    volt = joule/coulomb
    ampere = coulomb/second
    var = volt*ampere # reactive electric power

    ppm = 1.0

    # temperature difference
    F = scipy_cte.convert_temperature(2,'F','K') - scipy_cte.convert_temperature(1,'F','K')
    C = scipy_cte.convert_temperature(2,'C','K') - scipy_cte.convert_temperature(1,'C','K')
    K = 1.0
    kelvin = 1.0

    # temperature conversion
    # static method
    def convert_temperature(val, from_scale_name, to_scale_name):
        return scipy_cte.convert_temperature(val, from_scale_name, to_scale_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
"""Crimninal justice network dynamics modeling.

This example uses 7 modules:
    - Community
    - Arrested
    - Adjudication
    - Jail
    - Prison
    - Probation
    - Parole
and a population balance model is used to follow the offenders population
groups between modules.

To run this case using MPI you should compute the number of
processes as follows:

    `nprocs = 7 + 1 cortix`

then issue the MPI run command as follows (replace `nprocs` with a number):

     `mpiexec -n nprocs run_justice.py`

To run this case with the Python multiprocessing library, just run this file at the
command line as

    `run_city_justice.py`

To profile this run do:

     python -m cProfile -s time run_city_justice.py > lixo
"""

import scipy.constants as unit
import matplotlib.pyplot as plt

from cortix import Cortix
from cortix import Network

from cortix.examples.city_justice.prison import Prison
from cortix.examples.city_justice.parole import Parole
from cortix.examples.city_justice.adjudication import Adjudication
from cortix.examples.city_justice.jail import Jail
from cortix.examples.city_justice.arrested import Arrested
from cortix.examples.city_justice.probation import Probation
from cortix.examples.city_justice.community import Community

def main():
    '''Cortix run file for a criminal justice network.

    Attributes
    ----------
    n_groups : int
        Number of population groups being followed. This must be the same for all
        modules.
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    '''

    # Configuration Parameters
    n_groups  =  10 # number of population groups
    end_time  =  30* unit.year
    time_step = 1.0 * unit.day

    use_mpi = False  # True for MPI; False for Python multiprocessing

    city = Cortix(use_mpi=use_mpi, splash=True)

    city.network = Network()

    community = Community(n_groups=n_groups, non_offender_adult_population=100000,
                          offender_pool_size=0)
    city.network.module(community)
    community.end_time = end_time
    community.time_step = time_step
    community.show_time = (True, 1*unit.year)
    community.save = True

    prison = Prison(n_groups=n_groups, pool_size=0.0)
    city.network.module(prison)
    prison.end_time = end_time
    prison.time_step = time_step
    prison.save = True

    parole = Parole(n_groups=n_groups)
    city.network.module(parole)
    parole.end_time = end_time
    parole.time_step = time_step
    parole.save = True

    adjudication = Adjudication(n_groups=n_groups)
    city.network.module(adjudication)
    adjudication.end_time = end_time
    adjudication.time_step = time_step
    adjudication.save = True

    jail = Jail(n_groups=n_groups)
    city.network.module(jail)
    jail.end_time = end_time
    jail.time_step = time_step
    jail.save = True

    arrested = Arrested(n_groups=n_groups)
    city.network.module(arrested)
    arrested.end_time = end_time
    arrested.time_step = time_step
    arrested.save = True

    probation = Probation(n_groups=n_groups)
    city.network.module(probation)
    probation.end_time = end_time
    probation.time_step = time_step
    probation.save = True

    city.network.connect(prison, parole, 'bidirectional')
    city.network.connect(adjudication, prison, 'directional')
    city.network.connect(jail, prison, 'directional')
    city.network.connect(adjudication, jail, 'directional')
    city.network.connect(arrested, jail, 'directional')
    city.network.connect(arrested, adjudication, 'directional')
    city.network.connect(arrested, probation, 'directional')
    city.network.connect(probation, jail, 'directional')
    city.network.connect(adjudication, probation, 'directional')

    city.network.connect(arrested, community, 'bidirectional')
    city.network.connect(jail, community, 'directional')
    city.network.connect(probation, community, 'directional')
    city.network.connect(adjudication, community, 'directional')
    city.network.connect(prison, community, 'directional')
    city.network.connect(parole, community, 'directional')

    city.network.draw()
    #city.network.draw(engine='dot')

    city.run()

    if city.use_multiprocessing or city.rank == 0:

        total_num_unknowns = n_groups * len(city.network.modules)
        total_num_params = 0

        # Inspect Data Function
        def inspect_module_data(module, quant_name):
            population_phase = module.population_phase
            (fxg_quant, time_unit) = population_phase.get_quantity_history(quant_name)

            fxg_quant.plot(x_scaling=1/unit.year, x_label='Time [y]',
                           y_label=fxg_quant.latex_name+' ['+fxg_quant.unit+']')

            # Number of parameters in the prison model
            n_params = (len(population_phase.actors)-1)*n_groups
            return n_params

        quant_names = {'Prison':'npg','Parole':'feg','Adjudication':'fag',
                'Jail':'fjg','Arrested':'frg','Probation':'fbg','Community':'f0g'}

        for m in city.network.modules:
            quant_name = quant_names[m.name]
            total_num_params += inspect_module_data(m,quant_name)
            plt.grid()
            plt.savefig(m.name+'.png', dpi=300)
            if m.name=='Community':
                inspect_module_data(m,'n0')
                plt.grid()
                plt.savefig(m.name+'-n0.png', dpi=300)

        # Total number of unknowns and parameters

        city.log.info('total number of unknowns   ='+str(total_num_unknowns))
        city.log.info('total number of parameters ='+str(total_num_params))

    # Properly shutdow city
    city.close()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
import logging

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class Community(Module):
    """Community model with criminal group population.

    Community here is the system at large with all possible
    adult individuals included in any given society.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules:

     + `probation`,
     + `adjudication`,
     + `jail`, `prison`,
     + `arrested`, and
     + `parole`.

    See instance attribute `port_names_expected`.

    """

    def __init__(self, n_groups=1, non_offender_adult_population=100,
                 offender_pool_size=0.0, free_offender_pool_size=0.0):
        """Constructor.

        Parameters
        ----------
        n_groups: int
            Number of groups in the population.
        non_offender_adult_population: float
            Pool of individuals reaching the adult age (SI) unit. Default: 100.
        offender_pool_size: float
            Upperbound on the range of the existing population groups. A random value
            from 0 to the upperbound value will be assigned to each group. This is
            typically a small number, say a fraction of a percent.
        """

        super().__init__()

        self.port_names_expected = ['probation', 'adjudication', 'jail', 'prison',
                                    'arrested', 'parole']

        quantities = list()
        self.ode_params = dict()

        self.initial_time = 0.0 * unit.day
        self.end_time = 100 * unit.day
        self.time_step = 0.5 * unit.day
        self.show_time = (False, 10*unit.day)
        self.log = logging.getLogger('cortix')

        unit.percent = 1/100

        # Population groups
        self.n_groups = n_groups

        # Community non-offender population
        n0_0 = float(non_offender_adult_population)
        n0 = Quantity(name='n0', formal_name='non-offender-adult-pop',
                      latex_name='$n_0$',
                      unit='# adults', value=n0_0,
                      info='Non-Offender Adult Population')
        quantities.append(n0)

        # Community free-offender population groups
        f0g_0 = np.random.random(self.n_groups) * offender_pool_size
        f0g = Quantity(name='f0g', formal_name='free-offender-pop-grps',
                       latex_name='$n_0^{(g)}$',
                       unit='# offenders', value=f0g_0,
                       info='Free-Offender Population Groups')
        quantities.append(f0g)

        # Model parameters: commitment coefficients

        # Community non-offenders to offenders (arrested)
        a = 0.6*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        c00g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        c00g = Quantity(name='c00g',
                        formal_name='non-offenders-commit-arrested-coeff-grps',
                        unit='1/s', value=c00g_0)
        self.ode_params['non-offenders-commit-to-arrested-coeff-grps'] = c00g_0
        quantities.append(c00g)

        # Community free-offenders to arrested (recidivism)
        a = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.9*unit.percent/unit.year * np.ones(self.n_groups)
        c0rg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        c0rg = Quantity(name='c0rg',
                        formal_name='free-offenders-commit-arrested-coeff-grps',
                        value=c0rg_0, unit='1/s')
        self.ode_params['free-offenders-commit-to-arrested-coeff-grps'] = c0rg_0
        quantities.append(c0rg)

        # Death term for community offenders
        a = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        b = 1.0*unit.percent/unit.year * np.ones(self.n_groups)
        d0g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        self.ode_params['free-offenders-death-rates-coeff'] = d0g_0

        # Death term for community non-offenders
        a = 0.5*unit.percent/unit.year
        b = 1.0*unit.percent/unit.year
        d0_0 = a + (b-a)*np.random.random()
        self.ode_params['non-offenders-death-rate-coeff'] = d0_0

        # Maturity term for community non-offenders
        a = 1.5*unit.percent/unit.year
        b = 2.5*unit.percent/unit.year
        s0_0 = a + (b-a)*np.random.random()
        self.ode_params['non-offenders-maturity-rate-coeff'] = s0_0
        self.ode_params['non_offender_adult_population'] = non_offender_adult_population

        # Phase state
        self.population_phase = Phase(self.initial_time, time_unit='s',
                                      quantities=quantities)

        self.population_phase.set_value('n0', n0_0, self.initial_time)
        self.population_phase.set_value('f0g', f0g_0, self.initial_time)

        # Initialize inflows to zero
        self.ode_params['prison-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['parole-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['arrested-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['jail-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['adjudication-inflow-rates'] = np.zeros(self.n_groups)
        self.ode_params['probation-inflow-rates'] = np.zeros(self.n_groups)

    def run(self, *args):

        # Some logic for logging time stamps
        # Leave this here: rebuild logger
        logger_name = args[0][0].name
        self.rebuild_logger(logger_name)

        self.__zero_ode_parameters()

        time = self.initial_time

        while time <= self.end_time:

            if self.show_time[0] and abs(time%self.show_time[1]-0.0) <= 1.e-1:
                self.log.info('Community::time[y] = '+str(round(time/unit.year, 1)))

            self.__call_ports(time)

            # Evolve offenders group population to the next time stamp
            #---------------------------------------------------------

            time = self.__step(time)

    def __call_ports(self, time):

        # Interactions in the jail port
        #--------------------------------
        # one way "from" jail

        self.send(time, 'jail')
        (check_time, inflow_rates) = self.recv('jail')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['jail-inflow-rates'] = inflow_rates

        # Interactions in the adjudication port
        #--------------------------------------
        # one way "from" adjudication

        self.send(time, 'adjudication')
        (check_time, inflow_rates) = self.recv('adjudication')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['adjudication-inflow-rates'] = inflow_rates

        # Interactions in the probation port
        #--------------------------------
        # one way "from" probation

        self.send(time, 'probation')
        (check_time, inflow_rates) = self.recv('probation')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['probation-inflow-rates'] = inflow_rates

        # Interactions in the prison port
        #--------------------------------
        # one way "from" prison

        self.send(time, 'prison')
        (check_time, inflow_rates) = self.recv('prison')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['prison-inflow-rates'] = inflow_rates

        # Interactions in the parole port
        #--------------------------------
        # one way "from" parole

        self.send(time, 'parole')
        (check_time, inflow_rates) = self.recv('parole')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['parole-inflow-rates'] = inflow_rates

        # Interactions in the arrested port
        #--------------------------------
        # two way "to" and "from" arrested

        # to
        message_time = self.recv('arrested')
        outflow_rates = self.__compute_outflow_rates(message_time, 'arrested')
        self.send((message_time, outflow_rates), 'arrested')

        # from
        self.send(time, 'arrested')
        (check_time, inflow_rates) = self.recv('arrested')
        assert abs(check_time-time) <= 1e-6
        self.ode_params['arrested-inflow-rates'] = inflow_rates

    def __step(self, time=0.0):
        r"""ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None
        """

        # Get state values
        a_val = self.population_phase.get_value('n0', time)
        b_vec = self.population_phase.get_value('f0g', time)

        u_0 = np.concatenate((np.array([a_val]), b_vec))

        t_interval_sec = np.linspace(0.0, self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                         u_0, t_interval_sec,
                                         args=(self.ode_params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1, :]  # solution vector at final time step


        time += self.time_step

        # Update state variables
        values = self.population_phase.get_row() # values existing values
        self.population_phase.add_row(time, values) # copy on new time for convenience

        self.population_phase.set_value('n0', u_vec[:1], time) # insert new values
        self.population_phase.set_value('f0g', u_vec[1:], time) # insert new values

        return time

    def __rhs_fn(self, u_vec, t, params):

        n0 = u_vec[:1]   # non-offender population

        # source of non-offenders
        s0_coeff = params['non-offenders-maturity-rate-coeff']
        n0_0 = params['non_offender_adult_population']
        s0 = s0_coeff * n0_0

        # outflow rate to Arrested
        c00g = params['non-offenders-commit-to-arrested-coeff-grps']
        outflow_rate = np.sum(c00g * n0)

        # death rate
        death_rate_coeff = params['non-offenders-death-rate-coeff']
        death_rate = death_rate_coeff * n0

        dt_n0 = s0 - outflow_rate - death_rate


        f0g = u_vec[1:]  # free-offender population groups

        prison_inflow_rates = params['prison-inflow-rates']
        parole_inflow_rates = params['parole-inflow-rates']
        arrested_inflow_rates = params['arrested-inflow-rates']
        jail_inflow_rates = params['jail-inflow-rates']
        adjudication_inflow_rates = params['adjudication-inflow-rates']
        probation_inflow_rates = params['probation-inflow-rates']

        # free-offenders inflows
        inflow_rates = prison_inflow_rates + parole_inflow_rates +\
                       arrested_inflow_rates + jail_inflow_rates +\
                       adjudication_inflow_rates + probation_inflow_rates

        assert np.all(inflow_rates >= 0.0), 'values: %r'%inflow_rates

        c0rg = params['free-offenders-commit-to-arrested-coeff-grps']


        # free-offenders outflow (recidivism)
        outflow_rates = c0rg * f0g

        assert np.all(outflow_rates >= 0.0), 'values: %r'%outflow_rates

        death_rates_coeff = params['free-offenders-death-rates-coeff']

        death_rates = death_rates_coeff * f0g

        assert np.all(death_rates >= 0.0), 'values: %r'%death_rates

        dt_f0g = inflow_rates - outflow_rates - death_rates


        dt_u = np.concatenate((dt_n0, dt_f0g))

        return dt_u

    def __compute_outflow_rates(self, time, name):

        n0 = self.population_phase.get_value('n0', time)
        f0g = self.population_phase.get_value('f0g', time)

        if name == 'arrested':

            c0rg = self.ode_params['free-offenders-commit-to-arrested-coeff-grps']

            c00g = self.ode_params['non-offenders-commit-to-arrested-coeff-grps']

            # Recidivism and new offenders
            outflow_rates = c0rg * f0g + c00g * n0

        return outflow_rates

    def __zero_ode_parameters(self):
        '''
        If ports are not connected the corresponding outflows must be zero.

        '''

        zeros = np.zeros(self.n_groups)

        p_names = [p.name for p in self.ports]

        if 'arrested' not in p_names:
            self.ode_params['commit-to-arrested-coeff-grps'] = zeros

            self.ode_params['general-commit-to-arrested-coeff-grps'] = zeros

        return
